package Fedora::Rebuild::Package;
use strict;
use warnings;
use threads;
use threads::shared;
use Moose;
use File::Path;
use File::Spec;
use Fedora::Rebuild::Package::StateLock;
use Fedora::Rebuild::RPM;
use Data::Dumper;
use namespace::clean;

use version 0.77; our $VERSION = version->declare("v0.8.0");

has 'name' => (is => 'ro', isa => 'Str', required => 1);
# Build-time dependencies
has 'requires' => (is => 'rw', isa => 'HashRef', lazy_build => 1,
    init_arg => undef);
# Run-time dependencies hashed by binary package ENVR.
has 'runrequires' => (is => 'rw', isa => 'HashRef', lazy_build => 1,
    init_arg => undef);
# Run-time provides hashes by binary package ENVR.
has 'provides' => (is => 'rw', isa => 'HashRef', lazy_build => 1,
    init_arg => undef);
has 'workdir' => (is => 'ro', isa => 'Str', required => 1);
# Git branch name
# "f14", "f15" etc. Use "rawhide" for latest one.
has 'dist' => (is => 'ro', isa => 'Str', required => 1);
# Build target name
# "dist-f14", "dist-f15" etc. Use "dist-rawhide" for latest one.
has 'target' => (is => 'ro', isa => 'Str', required => 1);
has 'message' => (is => 'ro', isa => 'Str', required => 1);

has 'packagedir' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'repodir' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'rpmdir' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'requiresstore' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'runrequiresstore' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'providesstore' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'taskstore' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);
has 'branch' => (is => 'ro', isa => 'Str', lazy_build => 1,
    init_arg => undef);

# Make object shared between threads.
# XXX: Not all attributes are shared automatically.
around 'new' => sub {
    my $orig = shift;
    my $class = shift;
    return shared_clone($class->$orig(@_));
};

# Clean package locks bound to attributes
sub BUILD {
    my $self = shift;
    for my $state qw(buildrequires runrequires provides) {
        Fedora::Rebuild::Package::StateLock->new(package => $self,
            state => $state)->mark_failed;
    }
}

# BuildRequires. Shared.
# Initialize to empty hash. Call get_buildrequires() to populate.
sub _build_requires {
    my $self = shift;
    my $requires :shared = shared_clone({});
    return $requires;
}

# Run-time Requires. Shared.
# Initialize to empty hash. Call get_runrequires() to populate.
sub _build_runrequires {
    my $self = shift;
    my $runrequires :shared = shared_clone({});
    return $runrequires;
}

# Run-time provides. Shared.
# Initialize to empty hash. Call rebuild() or get_binaryprovides() to populate.
sub _build_provides {
    my $self = shift;
    my $provides :shared = shared_clone({});
    return $provides;
}

sub _build_packagedir {
    my $self = shift;
    my $directory = File::Spec->catfile($self->workdir, $self->name);
    if (! -d $directory) {
        File::Path::make_path($directory) or
            die "Could not create directory $directory: $!";
    }
    return $directory;
}

sub _build_repodir {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'repository');
}

sub _build_rpmdir {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'RPMS');
}

sub _build_requiresstore {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'buildrequires.store');
}

sub _build_runrequiresstore {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'runrequires.store');
}

sub _build_providesstore {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'provides.store');
}

sub _build_taskstore {
    my $self = shift;
    return File::Spec->catfile($self->packagedir, 'task.store');
}

sub _build_branch {
    my $self = shift;
    if ($self->dist eq 'rawhide') {
        return 'master';
    } else {
        return $self->dist;
    }
}


# Clones package repository and switch to proper branch if not yet done.
# Return true on success.
sub clone {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'clone');
    if ($lock->is_done) { return 1; }

    # XXX: fedpkg creates subdirectory named like package. The package name
    # could clash with our file structure. Thus we need to clone into secure
    # directory and move repository content to fixed name $self->repodir after
    # that.

    if (-d $self->repodir) { File::Path::remove_tree($self->repodir); }
    my $tempdir = File::Spec->catfile($self->packagedir, 'tmp');
    my $temprepodir = File::Spec->catfile($tempdir, $self->name);
    if (-d $tempdir) { File::Path::remove_tree($tempdir); }
    if (!File::Path::make_path($tempdir)) {
        $lock->log("Could not create directory `" . $tempdir . "': $!\n");
        return $lock->mark_failed;
    }

    if (!$lock->do($tempdir, 'fedpkg', 'clone', $self->name)) {
        $lock->log("Could not clone `" . $self->name . "' repository.\n");
        return $lock->mark_failed;
    }

    if (!rename($temprepodir, $self->repodir)) {
        $lock->log("Could not move `" . $temprepodir . "' content to to `" .
            $self->repodir . "'.\n");
        return $lock->mark_failed;
    }
    File::Path::remove_tree($tempdir);

    if (!$lock->do($self->repodir, 'fedpkg', 'switch-branch', $self->branch)) {
        $lock->log("Could not switch `" . $self->name .
            "' repository to branch `" . $self->branch . "'.\n");
        return $lock->mark_failed;
    }

    return $lock->mark_done;
}

# Builds SRPM locally. Return true on success.
# If first argument is true, recreate SRPM forcefully.
# Needs `clone'.
sub srpm {
    my ($self, $force) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'srpm');
    if ($force) { $lock->remove_lock; }
    if ($lock->is_done) { return 1; }

    if (!$lock->do($self->repodir, 'fedpkg', 'srpm')) {
        $lock->log("Could not build SRPM for `" . $self->name .
            "' package locally.\n");
        return $lock->mark_failed;
    }

    return $lock->mark_done;
}

# Get current package NEVRA from sources in repository.
# First argument is state lock where process of getting NEVRA including
# potential failure is logged.
# XXX: The state is not marked as failed in case of error,
# Return NEVRA string or undef in case of error.
sub get_nevra_from_git {
    my ($self, $lock) = @_;

    my $nevra;
    if (!$lock->dooutput($self->repodir, \$nevra, 'fedpkg', 'verrel') ||
            $nevra eq '') {
        $lock->log("Could not get NEVRA from `" . $self->name .
            "' git repository package.\n");
        return undef;
    }
    chomp $nevra;
    # Consider last line only becuase of bug in fedpkg
    # <https://bugzilla.redhat.com/show_bug.cgi?id=721389>.
    my @lines = (split qr{$/}, $nevra);
    $nevra = pop @lines;
    return $nevra;
}

# Destile BuildRequires from local SRPM and serialize them into file.
# Return true on success.
# Needs `srpm'.
# FIXME: does not work after cleaning clone or doing update.
sub storebuildrequires {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'buildrequiresstore');
    if ($lock->is_done) { return 1; }

    my $nevra = $self->get_nevra_from_git($lock);
    if (! defined $nevra) {
        return $lock->mark_failed;
    }

    my $srpmname = File::Spec->catfile($self->repodir, $nevra . '.src.rpm');
    if (! -f $srpmname ) {
        $lock->log("SRPM package `" . $srpmname . "' is missing, " .
            "trying to create SRPM again...\n");
        if (!$self->srpm(1) || ! -f $srpmname) {
            $lock->log("`Could not recreate SRPM package '" . $srpmname .
                "'.\n");
            return $lock->mark_failed;
        }
    }

    my $rpm = Fedora::Rebuild::RPM->new(name => $srpmname);
    my ($requires, $envra) = $rpm->requires;
    if (! defined $requires) {
        $lock->log("Could not get requires of SRPM `" . $srpmname . "': ". $@
            . "\n");
        return $lock->mark_failed;
    }

    if (! $lock->nstorereference($requires, $self->requiresstore)) {
        $lock->log("Could not store requires of SRPM `". $srpmname .
            "' into `" . $self->requiresstore . "' file: $@\n");
        return $lock->mark_failed;
    }

    $lock->log(Data::Dumper::Dumper($requires) . "\n");
    return $lock->mark_done;
}

# Destile BuildRequires from local SRPM. Return true on success.
# Needs `buildrequiresstore'.
sub buildrequires {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'buildrequires');
    if ($lock->is_done) { return 1; }

    my $requires = $lock->retrievereference($self->requiresstore);
    if (! $requires) {
        $lock->log("Could not load requires of `". $self->name .
            "' package from `" . $self->requiresstore . "' file: $@\n");
        return $lock->mark_failed;
    }
    $self->requires(shared_clone($requires));

    $lock->log(Data::Dumper::Dumper($self->requires) . "\n");
    return $lock->mark_done;
}

# Record verdict from dependency solver whether the package is rebuildable.
# This step is always redone.
sub log_is_rebuildable {
    my ($self, $is_rebuildable, $message) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'is_rebuildable');
    $lock->remove_lock;

    $lock->log("Solver result for possibility of rebuilding SRPM for `" .
        $self->name . "': $message\n");
    if (! $is_rebuildable) {
        $lock->log("According dependency solver, this package is not " .
            "rebuildable now.\n");
        return $lock->mark_failed;
    }

    $lock->log("According dependency solver, this package is " .
        "rebuildable now.\n");
    return $lock->mark_done;
}

# Get binary RPM packages for the source package.
# Download them from Koji, or from local build if first argument is true.
# Requires `clone'. Sould be called after `build'.
# Return true on success.
sub binaryrpm {
    my ($self, $local) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'rpms');
    if ($lock->is_done) { return 1; }

    if (-d $self->rpmdir) { File::Path::remove_tree($self->rpmdir); }
    if (!File::Path::make_path($self->rpmdir)) {
        $lock->log("Could not create directory `" . $self->rpmdir . "': $!\n");
        return $lock->mark_failed;
    }

    # TODO: Get current architecture by rpmGetArchInfo() from librpm
    my @archs = qw(x86_64 noarch);
    if (!$local) {
        $lock->log("Getting binary RPM packages from Koji:\n");

        my $nevra = $self->get_nevra_from_git($lock);
        if (! defined $nevra) {
            return $lock->mark_failed;
        }

        # TODO: Get all archs, remove SRPM
        if (!$lock->do($self->rpmdir, 'koji', 'download-build',
                (map {'--arch=' . $_ } @archs), $nevra)) {
            $lock->log("Could get binary RPM packages for `" . $nevra . "'\n");
            return $lock->mark_failed;
        }
    } else {
        $lock->log("Getting binary RPM packages from local build:\n");

        my @rpms = map { glob(File::Spec->catfile($self->repodir, $_,
            '*.rpm')) } @archs;
        if ($#rpms < 0) {
            $lock->log("No binary RPM packages found under `" .
                $self->repodir . "'\n");
            return $lock->mark_failed;
        }

        for my $rpm (@rpms) {
            use File::Copy;
            if (!copy($rpm, $self->rpmdir)) {
                $lock->log("Could not copy `$rpm' into `". $self->rpmdir .
                    "'\n");
                return $lock->mark_failed;
            }
            $lock->log("`" . $rpm . "' copied into `" . $self->rpmdir . "'\n");
        }
    }

    return $lock->mark_done;
}

# Distill Requires from rebuilt binary packages and serialize them into file.
# Return true on success.
# Needs `rpms'.
sub storebinaryrequires {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'runrequiresstore');
    if ($lock->is_done) { return 1; }

    my @rpms = glob(File::Spec->catfile($self->rpmdir, '*.rpm'));
    if ($#rpms < 0) {
        $lock->log("No binary RPM packages found in `" . $self->rpmdir
            . "'\n");
        return $lock->mark_failed;
    }

    my $allrequires = {};
    for my $rpmname (@rpms) {
        my $rpm = Fedora::Rebuild::RPM->new(name => $rpmname);

        my ($requires, $envr) = $rpm->requires;

        if (! defined $requires || ! defined $envr) {
            $lock->log("Could not get run-time requires of RPM `" . $rpmname .
                "': " .  $@ ."\n");
            return $lock->mark_failed;
        }
        $$allrequires{$envr} = $requires;
    }

    if (! $lock->nstorereference($allrequires, $self->runrequiresstore)) {
        $lock->log("Could not store run-time requires of RPM `". $self->name .
            "' into `" . $self->runrequiresstore . "' file: $@\n");
        return $lock->mark_failed;
    }

    $lock->log(Data::Dumper::Dumper($allrequires));
    return $lock->mark_done;
}

# Distill Provides from rebuilt binary packages and serialize them into file.
# Return true on success.
# Needs `rpms'.
sub storebinaryprovides {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'providesstore');
    if ($lock->is_done) { return 1; }

    my @rpms = glob(File::Spec->catfile($self->rpmdir, '*.rpm'));
    if ($#rpms < 0) {
        $lock->log("No binary RPM packages found in `" . $self->rpmdir
            . "'\n");
        return $lock->mark_failed;
    }

    my $allprovides = {};
    for my $rpmname (@rpms) {
        my $rpm = Fedora::Rebuild::RPM->new(name => $rpmname);

        my ($provides, $envr) = $rpm->provides;

        if (! defined $provides || !defined $envr) {
            $lock->log("Could not get provides of RPM `" . $rpmname . "': " .
                $@ ."\n");
            return $lock->mark_failed;
        }
        $$allprovides{$envr} = $provides;
    }

    if (! $lock->nstorereference($allprovides, $self->providesstore)) {
        $lock->log("Could not store provides of RPM `". $self->name .
            "' into `" . $self->providesstore . "' file: $@\n");
        return $lock->mark_failed;
    }

    $lock->log(Data::Dumper::Dumper($allprovides));
    return $lock->mark_done;
}

# Load run-time requires of already rebuilt binary packages from file.
# Return true on success.
# Needs `storebinaryrequires'.
sub loadbinaryrequires {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'runrequires');
    if ($lock->is_done) { return 1; }

    my $runrequires = $lock->retrievereference($self->runrequiresstore);
    if (! $runrequires) {
        $lock->log("Could not load run-time requires of `". $self->name .
            "' package from `" . $self->runrequiresstore . "' file: $@\n");
        return $lock->mark_failed;
    }
    $self->runrequires(shared_clone($runrequires));
    
    $lock->log(Data::Dumper::Dumper($self->runrequires));
    return $lock->mark_done;
}

# Load provides of already rebuilt binary packages from file.
# Return true on success.
# Needs `storebinaryprovides'.
sub loadbinaryprovides {
    my $self = shift;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'provides');
    if ($lock->is_done) { return 1; }

    my $provides = $lock->retrievereference($self->providesstore);
    if (! $provides) {
        $lock->log("Could not load provides of `". $self->name .
            "' package from `" . $self->providesstore . "' file: $@\n");
        return $lock->mark_failed;
    }
    $self->provides(shared_clone($provides));
    
    $lock->log(Data::Dumper::Dumper($self->provides));
    return $lock->mark_done;
}

# Increase package revision if not yet done. Commit change if first argument
# is not true.
# Return true on success.
# Needs `clone'.
sub update {
    my ($self, $local) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'update');
    if ($lock->is_done) { return 1; }

    # Reset git repository
    if (!$lock->do($self->repodir, 'git', 'reset', '--hard',
            'origin/' .  $self->branch)) {
        $lock->log("Could not reset git repository in `" . $self->repodir .
            "'.\n");
        return $lock->mark_failed;
    }
   
    # Pull git repository
    if (!$lock->do($self->repodir, 'git', 'pull')) {
        $lock->log("Could not pull git repository in `" . $self->repodir .
            "'.\n");
        return $lock->mark_failed;
    }
   
    # Increase revision number
    my $specfile = File::Spec->catfile($self->repodir, $self->name . '.spec');
    if (!$lock->do(undef, 'rpmdev-bumpspec', '-c', $self->message, $specfile)) {
        $lock->log("Could not increase revison number in `" . $specfile .
            "' specfile.\n");
        return $lock->mark_failed;
    }

    # Commit changes
    if (!$lock->do($self->repodir, 'git', 'commit', '-a',
            '-m', $self->message)) {
        $lock->log("Could not commit changes into git repository `" .
            $self->repodir . "'.\n");
        return $lock->mark_failed;
    }
   
    if ($local) {
        $lock->log("Not pushing changes because of local build mode.\n");
    } else {
        # Push changes
        if (!$lock->do($self->repodir, 'git', 'push')) {
            $lock->log("Could not push changes from repository `" .
                $self->repodir . "' to server.\n");
            return $lock->mark_failed;
        }
    }
   
    return $lock->mark_done;
}


# Submit package for building into Koji and store task ID.
# This is pointless in local build mode.
# Requires `clone'. Sould be called after `update'.
# Return true on success.
sub submitbuild {
    my ($self) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'submitbuild');
    if ($lock->is_done) { return 1; }

    # Get NEVRA of intended build
    my $nevra = $self->get_nevra_from_git($lock);
    if (! defined $nevra) {
        return $lock->mark_failed;
    }

    # Check the build is not already in Koji (e.g. by concurrent packager
    # or after this program restart) but do not conclude anything.
    my $buildinfo;
    if (!$lock->dooutput($self->repodir, \$buildinfo, 'koji', 'buildinfo',
            $nevra)) {
        $lock->log("Could not ask Koji for `" . $nevra . "' status " .
            "before submitting new build.\n");
        return $lock->mark_failed;
    }
    if ($buildinfo =~ /No such build/m) {
        $lock->log("Package not yet submitted for building as expected.\n");
    } else {
        # Get task ID of already building package
        if ($buildinfo =~ /Task:\s*(\d+)/m) {
            # TODO: We could compare task target and consider as submitted if
            # equaled to intended target.
            my $task_id = $1;
            $lock->log("Package `$nevra' already submitted as task " .
                "`$task_id'. Previous build failed or somebody builds the " .
                "package concurrently.\n");
        } else {
            $lock->log("Package `$nevra' already in Koji, but task ID " .
                "could not been determined.\n");
        }
        $lock->log("Re-submitting the package.\n")
    }

    # Submit into Koji
    my $task_id;
    if (!$lock->dooutput($self->repodir, \$task_id, 'fedpkg', 'build',
            '--nowait', '--target', $self->target)) {
        $lock->log("Could not submit `" . $nevra . "' into Koji.\n");
        return $lock->mark_failed;
    }
    if (not $task_id =~ /Created task:\s*(\d+)/) {
        $lock->log("Could not parse Koji task ID for `$nevra' build\n");
        return $lock->mark_failed;
    }
    $task_id = $1;

    # Store task ID
    if (! $lock->nstorereference(\$task_id, $self->taskstore)) {
        $lock->log("Could not store task ID `" . $task_id . "' of `" . $nevra .
            "' package into `" . $self->taskstore . "' file: $@\n");
        return $lock->mark_failed;
    } else {
        $lock->log("Task ID `" . $task_id . "' stored into `" .
            $self->taskstore . "' file sucessfully.\n");
    }

    return $lock->mark_done;
}


# Wait for package build in Koji, or if first argument is true, build locally
# Requires `clone'. Sould be called after `update' or `submitbuild'.
# Return true on success.
sub build {
    my ($self, $local) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'build');
    if ($lock->is_done) { return 1; }

    my $nevra = $self->get_nevra_from_git($lock);
    if (! defined $nevra) {
        return $lock->mark_failed;
    }

    if ($local) {
        if (!$lock->do($self->repodir, 'fedpkg', 'local')) {
            $lock->log("Could not build `" . $nevra . "' locally.\n");
            return $lock->mark_failed;
        }
    } else {
        # Retrieve task ID of submitted build
        my $task_id = $lock->retrievereference($self->taskstore);
        if (! $task_id) {
            $lock->log("Could not load task ID for `". $nevra .
                "' build from `" . $self->taskstore . "' file: $@\n");
            return $lock->mark_failed;
        }
        $task_id = $$task_id;
       
        # Wait for build task result
        # TODO: How to recognize the process died for other reason
        # than build failure?
        if (!$lock->do($self->repodir, 'koji', 'watch-task', $task_id)) {
            $lock->log("Could not get status of Koji task `" . $task_id .
                "' for `$nevra' build, or the task failed.\n");
            return $lock->mark_failed;
        }
    }

    return $lock->mark_done;
}


# Waits for Koji build root rotation. Just built package will be available for
# other packages at build time after returning from this fuction.
# If first argument is true, build is local and this function is void,
# otherwise waits for the Koji rotation.
# Requires `update'. Sould be called after `build'.
# Return true on success.
sub dowaitforbuildroot {
    my ($self, $local) = @_;
    my $lock = Fedora::Rebuild::Package::StateLock->new(package => $self,
       state => 'rotate');
    if ($lock->is_done) { return 1; }

    my $nevra = $self->get_nevra_from_git($lock);
    if (! defined $nevra) {
        return $lock->mark_failed;
    }

    if ($local) {
        $lock->log("`" . $nevra .
            "' built locally, not waiting on Koji rotation.\n");
        return $lock->mark_done;
    } else {
        if (!$lock->do($self->repodir, 'koji', 'wait-repo', 
                '--build=' . $nevra, '--target', $self->target)) {
            $lock->log("Koji does not contain `" . $nevra .
                "' package in build root for `" . $self->target .
                "' build target yet.\n");
            return $lock->mark_failed;
        }
    }

    return $lock->mark_done;
}

# Set hash of build-time dependencies (requires attribute).
# Return undef in case of failure.
sub get_buildrequires {
    my $self = shift;
    my $ok;
    print "Getting BuildRequires for `" . $self->name . "'...\n";

    # Procede all steps, each must be re-doable
    $ok = $self->clone;
    $ok = $self->srpm if $ok;
    $ok = $self->storebuildrequires if $ok;
    $ok = $self->buildrequires if $ok;
    
    #...;

    if ($ok) {
        print "BuildRequires for `" . $self->name .
            "' package distilled successfully.\n";
        return 1;
    } else {
        print "Could not get BuildRequires for `" . $self->name .
            "' package.\n";
        return undef;
    }
}

# Set hash of run-time requires (Requires attribute).
# Return true on success, undef in case of failure.
# XXX: Requires `runrequiresstore'
sub get_binaryrequires {
    my $self = shift;
    my $ok;
    print "Getting run-time requires for `" . $self->name . "'...\n";

    $ok = $self->storebinaryrequires;
    $ok = $self->loadbinaryrequires if $ok;

    if ($ok) {
        print "Run-time requires for `" . $self->name .
            "' package distilled successfully.\n";
        return 1;
    } else {
        print "Could not get run-time requires for `" . $self->name .
            "' package.\n";
        return undef;
    }
}

# Set hash of run-time provides (provides attribute).
# Return true on success, undef in case of failure.
# XXX: Requires `providesstore'
sub get_binaryprovides {
    my $self = shift;
    my $ok;
    print "Getting binary provides for `" . $self->name . "'...\n";

    $ok = $self->storebinaryprovides;
    $ok = $self->loadbinaryprovides if $ok;

    if ($ok) {
        print "Provides for `" . $self->name .
            "' package distilled successfully.\n";
        return 1;
    } else {
        print "Could not get Provides for `" . $self->name . "' package.\n";
        return undef;
    }
}

# Set hash of run-time requires (Requires attribute) and provides
# (provides attribute). Common wrapper for getbinaryrequires() and
# getbinaryprovides().
# Return true on success, undef in case of failure.
# XXX: Requires `runrequiresstore' and `providesstore'
sub get_binarydependencies {
    my $self = shift;
    my $ok;
    print "Getting binary dependencies for `" . $self->name . "'...\n";

    $ok = $self->get_binaryrequires;
    $ok = $self->get_binaryprovides if $ok;

    if ($ok) {
        print "Binary dependencies for `" . $self->name .
            "' package distilled successfully.\n";
        return 1;
    } else {
        print "Could not get binary dependencies for `" . $self->name .
            "' package.\n";
        return undef;
    }
}

# Rebuild a package without waiting for propagation to next build root.
# If first argument is true, build locally without pushing commits to server.
# Otherwise publish commit and build in Koji.
sub rebuild {
    my ($self, $local) = @_;
    my $ok;
    print "Rebuilding `" . $self->name . "'...\n";

    # Procede all steps, each must be re-doable
    $ok = $self->clone;
    $ok = $self->update($local) if $ok;
    $ok = $self->submitbuild($local) if ($ok and !$local);
    $ok = $self->build($local) if $ok;
    $ok = $self->binaryrpm($local) if $ok;
    $ok = $self->get_binaryprovides if $ok;
    $ok = $self->get_binaryrequires if $ok;

    if ($ok) {
        print "`" . $self->name . "' rebuild finished successfully.\n";
    } else {
        print "`" . $self->name . "' rebuild failed.\n";
    }
    return $ok;
}

# Wait for the package propagated into new build root.
# If first argument is true, waits locally (no-op).
# Otherwise waits for Koji build root rotation.
# XXX: Requires `update', should be called after rebuilding package.
sub waitforbuildroot {
    my ($self, $local) = @_;
    my $ok;
    print "Waiting for `" . $self->name . "' to get to build root...\n";

    $ok = $self->dowaitforbuildroot($local);

    if ($ok) {
        print "`" . $self->name . "' propagated successfully.\n";
    } else {
        print "`" . $self->name . "' propagation failed.\n";
    }
    return $ok;
}

1;
