package Fedora::Rebuild::Solver;
use strict;
use warnings;
use version 0.77; our $VERSION = version->declare("v0.8.0");

use Moose;
use Fedora::Rebuild::Set::Package;
use Fedora::Rebuild::Package;
use namespace::clean;

# Set of packages with populated provides and requires
has 'packages' => (is => 'ro', isa => 'Fedora::Rebuild::Set::Package',
    lazy_build => 1, required => 1);
# Reference to function with three arguments (dependency name, relation
# flag, version) returning false if the dependency should be considered, true
# otherwise. If attribute is undef or missing, no filtering will be performed
# (i.e. the same effect as sub {1}).
has 'dependencyfilter' => (is => 'ro', isa => 'CodeRef', required => 0);


# This is cache of binary packages already resolved as installable or not
# installable.
# To keep coherency, do not update $self->packages or referenced packages.
# This is reference to hash of binary package names. Values holds verdict
# (true for installable, false otherwise) and message (string of dependencies
# or reason why not installable).
has 'cache' => (is => 'rw', isa => 'HashRef', lazy_build => 1,
    init_arg => 0);

sub _build_cache {
    # TODO: Thread safety?
    return {};
}

sub invalidate_cache {
    my $self = shift;
    # TODO: Thread safety?
    $self->cache({});
    #print STDERR "Cache invalidated\n";
}

# Add is_installable result to cache
sub add_to_cache {
    my ($self, $binarypackage, $is_installable, $message) = @_;

    # TODO: Thread safety?
    ${$self->cache}{$binarypackage} = {
        is_installable => $is_installable,
        message => $message
    };

    #print STDERR "Cache write: $binarypackage, " .
    #    "is_installable=$is_installable, message=$message\n"; 
}

# Decide a package is installable now and set string of dependenices.
# This is private method, use is_installable() instead.
# $package is source package object
# $binarypackage is binary package ENVR as index to $package->runrequies
# $backpath is reference to list of binary packages on the path in the
# dependecy tree from parent to root (i.e. root is last element). It's used to
# detect cycle.
# Return 0 for false, 1 for true, 2 for cycle. Fill list of dependencies into
# $$message (without trailing new line) which is reference itself.
sub is_installable_recursive {

    my ($self, $package, $binarypackage, $message, $backpath) = @_;

    # Detect cycle
    if (defined $backpath) {
        for my $predecessor (@$backpath) {
            if ($predecessor eq $binarypackage) {
                return 2;
            }
        }
    } else {
        # Although the cache provides correct verdicts (installable vs. not
        # installable), it provides bad messages that were cached for
        # dependency in cycle.
        # Thus invalidate the cache for each top-level package
        $self->invalidate_cache();
    }
    unshift @$backpath, $binarypackage;

    # Check cached result
    if (defined (my $result = ${$self->cache()}{$binarypackage})) {
        $$message = $$result{message};
        #print STDERR "Cache hit: $binarypackage, message=$$message\n"; 
        shift @$backpath;
        return $$result{is_installable};
    };


    # Each requirement of $binarypackage must be satisfied by at least one
    # binary package from packages list.
    my $binarypackage_runrequires = ${$package->runrequires}{$binarypackage};
    for my $rname (keys %$binarypackage_runrequires) {
        for my $require (@{$$binarypackage_runrequires{$rname}}) {
            my $rflag = $$require[0];
            my $rversion = $$require[1];
            my $is_satisfied = 0;
            my $deep_message;

            if (defined $self->dependencyfilter and
                    &{$self->dependencyfilter}($rname, $rflag, $rversion)) {
                # Avoid user-defined Requires from dependency evaluation
                next;
            }

            ALL_PACKAGES: for my $providing_package ($self->packages->packages)
            {
                for my $binary_providing_package
                        (keys %{$providing_package->provides}) {
                    if (Fedora::Rebuild::RPM::is_satisfied($rname, $rflag,
                           $rversion,
                           ${$providing_package->provides}{$binary_providing_package})) {
                        $deep_message = '';
                        my $is_installable =
                                $self->is_installable_recursive(
                                    $providing_package,
                                    $binary_providing_package, \$deep_message,
                                    $backpath);
                        if ($is_installable) {
                            $is_satisfied = 1;

                            #print STDERR "$binary_providing_package: " .
                            #    "$is_installable, deep: $deep_message\n";
                            if ($is_installable == 1) {
                                # Do not log end of cycle
                                if ($$message ne '') {
                                    $$message .= ' ';
                                }
                                $$message .= $providing_package->name . '/' .
                                    $binary_providing_package;
                                if ($deep_message ne '')  {
                                    $$message .= " ($deep_message)";
                                }
                            }

                            last ALL_PACKAGES;
                        }
                    }
                }
            }

            if (!$is_satisfied) {
                $$message =  q{`} . $package->name . q{/} . $binarypackage .
                    q{' requires `}. $rname .
                    q{ } .
                    Fedora::Rebuild::RPM::flag_as_string($rflag) .
                    q{ } . $rversion . q{' that cannot be satisfied};
                if ($deep_message//'' ne '')  {
                    $$message .= " ($deep_message)";
                }

                # FIXME: Can we cache negative answers for cut-off subtree?
                $self->add_to_cache($binarypackage, 0, $$message);
                shift @$backpath;
                return 0;
            }
        }
    }

    # FIXME: We cannot cache package whose descendants were cut
    # off because of breaking cycle.
    # TODO: Cache top-level package only or reevaluate subtree
    # from point of view of the package. (Cyclic problem again?)
    $self->add_to_cache($binarypackage, 1, $$message);
    shift @$backpath;
    return 1; 
}


# Decides a $binarypackage compiled from source $package is installable now.
# This is hard problem as it requires recursive evaluation of each package
# satisifying run-time dependency of previous package.
# Return 0 for false, 1 or true, undef for error while deciding. Fill error
# message (without trailing new line) into reference of second argument
# if defined.
sub is_installable {
    my ($self, $package, $binarypackage, $message) = @_;

    my $reason = '';
    if ($self->is_installable_recursive($package, $binarypackage, \$reason,
            undef)) {
        if (defined $message) {
            $$message = q{RunRequires for `} . $package->name . '/' .
                $binarypackage . q{' fulfilled} .
                (($reason ne '') ? " ($reason)" : '') . '.';
        }
        return 1;
    } else {
        if (defined $message) {
            $$message = q{Package `} . $package->name . '/' . $binarypackage .
                q{' cannot be installed because: } . $reason . '.';
        }
        return 0;
    }
}


# Decide a package is buildable now and set string of dependenices.
# Return 0 for false, 1 or true. Fill list of dependencies into
# message (without trailing new line) which is reference itself.
sub is_buildable {

    my ($self, $package, $message) = @_;

    if (defined $message) {
        $$message = '';
    }

    # Each requirement of $package must be satisfied by at least one binary
    # package from source packages list.
    for my $rname (keys %{$package->requires}) {
        for my $require (@{${$package->requires}{$rname}}) {
            my $rflag = $$require[0];
            my $rversion = $$require[1];
            my $is_satisfied = 0;
            my $deep_message;

            if (defined $self->dependencyfilter and
                    &{$self->dependencyfilter}($rname, $rflag, $rversion)) {
                # Avoid user-defined BuilRequires from dependency evaluation
                next;
            }

            ALL_PACKAGES: for my $providing_package ($self->packages->packages)
            {
                for my $binary_providing_package
                        (keys %{$providing_package->provides}) {
                    if (Fedora::Rebuild::RPM::is_satisfied($rname, $rflag,
                            $rversion,
                            ${$providing_package->provides}{$binary_providing_package})) {
                        $deep_message = '';
                        if ($self->is_installable_recursive($providing_package,
                                $binary_providing_package, \$deep_message,
                                undef)) {
                            $is_satisfied = 1;
                            if ($message) {
                                if ($$message ne '') {
                                    $$message .= ' ';
                                }
                                $$message .= $providing_package->name . '/' .
                                    $binary_providing_package;
                                if ($deep_message ne '')  {
                                    $$message .= " ($deep_message)";
                                }
                            }
                            last ALL_PACKAGES;
                        }
                    }
                }
            }

            if (!$is_satisfied) {
                if (defined $message) {
                    $$message =  q{Source package `} . $package->name .
                        q{' cannot be installed because it build-requires `} .
                        $rname .  q{ } .
                        Fedora::Rebuild::RPM::flag_as_string($rflag) .
                        q{ } . $rversion . q{' that cannot be satisfied};
                    if ($deep_message//'' ne '')  {
                        $$message .= " ($deep_message)";
                    }
                    $$message .= '.';
                }
                return 0;
            }
        }
    }

    if (defined $message) {
        $$message = "BuildRequires for `" . $package->name . "' fulfilled" .
            (($$message ne '') ? " ($$message)" : '') . '.';
    }
    return 1; 
}


1;
